//----------------------------------------------------------------
// cmmsock.cc: Socket Interface for Cmm
//						(c) Ono, Tetsuo
//
//	Socket, GetHostName, GetHostAddress,
//	GetServerPort,
//	Bind, Connect, Accept, Listen
//----------------------------------------------------------------
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <cmm/cmm.h>
#include <cmm/cmmsock.h>

#if PLATFORM == FREEBSD
#define UNIX_PATH_MAX 104
#endif

static void MakeInetAddress (Value *addrp, Value *portp, struct sockaddr_in *sockaddrp)
{
    unsigned long in_addr ;
    
    sockaddrp->sin_family = AF_INET ;
    in_addr = ((addrp->arr.arrp[0].intval) & 0xff)  << 24 ;
    in_addr |= ((addrp->arr.arrp[1].intval) & 0xff) << 16 ;
    in_addr |= ((addrp->arr.arrp[2].intval) & 0xff) <<  8 ;
    in_addr |= ((addrp->arr.arrp[3].intval) & 0xff) ;
    sockaddrp->sin_addr.s_addr = htonl (in_addr) ;
    sockaddrp->sin_port   = htons (portp->intval) ;
}

//----------------------------------------------------------------
// GetHostName ()
//	description:
//		get the hostname of current machine
//	parameters:
//		none
//	returns:
// 		[success] name of machine
//		[failure] void
//----------------------------------------------------------------
static int __GetHostName (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    char namebuf[512] ;
    int len ;
    retval->Init () ;

    len = gethostname (namebuf, 512) ;
    if (len >= 512 || len < 0)
	return 0 ;
    char *s = cenv->heap_area.AllocString (namebuf, &cenv->varstack) ;
    *retval = s ;
    return 0 ;
}

//----------------------------------------------------------------
// GetHostAddress (hostname)
//	description:
//		returns an Ip address of specified host
//	parameters:
//		hostname ... name of host (string)
//	returns:
//		[success] ip address as a Cmm array(example [192, 3, 2, 4]).
//		[failure] void
//----------------------------------------------------------------
static int __GetHostAddress (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 1) {
	return 0 ;
    }
    if (!arg[0].IsString ()) {
	return 0 ;
    }
    struct hostent *hent ;
    hent = gethostbyname (arg[0].sptr) ;
    if (!hent)
	return 0 ;
    Value addrp ;
    addrp = cenv->AllocArray (4) ;
    addrp.arr.arrp[0] = ((hent->h_addr_list[0])[0]) & 0xff ;
    addrp.arr.arrp[1] = ((hent->h_addr_list[0])[1]) & 0xff ;
    addrp.arr.arrp[2] = ((hent->h_addr_list[0])[2]) & 0xff ;
    addrp.arr.arrp[3] = ((hent->h_addr_list[0])[3]) & 0xff ;
    *retval = addrp ;
    return 0 ;
}

//----------------------------------------------------------------
// Socket (family, type, protocol) ;
//	description:
//		create a socket.
//	parameters:
//		family ... An address family. Symbol AF_UNIX and AF_INET
//			   are currently available.(integer)
//		type ..... Communication type. SOCK_STREAM and SOCK_DGRAM
//			   are currently available.(integer)
//		protocol .. protocol usally set 0.(integer)
//	returns:
//		[success] a file descriptor of created socket(integer).
//		[failure] -1.
//----------------------------------------------------------------
static int __Socket (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    if (args != 3)
	return 0 ;
    if (!arg[0].IsInteger() || !arg[1].IsInteger() || !arg[2].IsInteger())
	return 0 ;
    int desc ;
    desc = socket (arg[0].intval, arg[1].intval, arg[2].intval) ;
    *retval = desc ;
    return 0 ;
}

//----------------------------------------------------------------
// Bind (socket, address, port)	 for INET domain 
// Bind (socket, unix_path) 	 for UNIX domain
//	description:
//		bind an address to the specified socket.
//	parameters:
//		socket .... a descriptor of a socket   (integer)
//		address ... ip address as a Cmm array  (array).
//		port ...... port number for the service(integer)
//		unix_path . socket name for a unix domain ipc (string).
//	returns:
//		[success] 0
//		[failure] -1
//----------------------------------------------------------------
static int __Bind (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    int st ;
    struct sockaddr_in inet_address ;
    struct sockaddr_un unix_domain_address ;
    struct sockaddr *addrp ;
    int addrlen ;

    *retval = -1 ;
    if (args != 3 && args != 2)
	return 0 ;
    if (!arg[0].IsInteger())
	return 0 ;
    if (arg[1].IsArray()) {
	if (args != 3)
	    return 0 ;
	if (!arg[2].IsInteger())
	    return 0 ;
	MakeInetAddress (&arg[1], &arg[2], &inet_address) ;
	addrp = (struct sockaddr *)&inet_address ;
	addrlen = sizeof (struct sockaddr_in) ;
    } else if (arg[1].IsString()) {
	if (args != 2)
	    return 0 ;
	if (strlen (arg[1].sptr) + 1 >= UNIX_PATH_MAX) {
	    return 0 ;
	}
	unix_domain_address.sun_family = AF_UNIX ;
	strcpy (unix_domain_address.sun_path, arg[1].sptr) ;
	addrp = (struct sockaddr *)&unix_domain_address ;
	addrlen = sizeof (struct sockaddr_un) ;
    } else {
	fprintf (stderr, "Bind: argument #2 invalid value type\n") ;
	return -1 ;
    }
    st = bind (arg[0].intval, addrp, addrlen) ;
    if (st < 0) {
	perror ("Bind: ") ;
	return 0 ;
    }
    *retval = 0 ;
    return 0 ;
}

//----------------------------------------------------------------
// Connect (socket, address, port)	for INET domain
// Connect (socket, unix_path)		for UNIX domain
//	description:
//		connect a specified socket to another socket.
//	parameters:
//		socket ..... a descriptor of a socket.(string)
//		address ... ip address as a Cmm array (array).
//		port ...... port number to connect to (integer).
//		unix_path . socket name to connect to (string).
//	returns:
//		[success] 0
//		[failure] -1
//----------------------------------------------------------------
static int __Connect (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    int st ;
    struct sockaddr_in inet_address ;
    struct sockaddr_un unix_domain_address ;
    struct sockaddr *addrp ;
    int addrlen ;

    *retval = -1 ;
    if (args != 2 && args != 3)
	return 0 ;
    if (!arg[0].IsInteger())
	return 0 ;
    if (arg[1].IsArray()) {
	if (args != 3)
	    return 0 ;
	if (!arg[2].IsInteger())
	    return 0 ;
	MakeInetAddress (&arg[1], &arg[2], &inet_address) ;
	addrp = (struct sockaddr *)&inet_address ;
	addrlen = sizeof (struct sockaddr_in) ;
    } else if (arg[1].IsString()) {
	if (args != 2)
	    return 0 ;
	if (strlen (arg[1].sptr) + 1 >= UNIX_PATH_MAX) {
	    return 0 ;
	}
	unix_domain_address.sun_family = AF_UNIX ;
	strcpy (unix_domain_address.sun_path, arg[1].sptr) ;
	addrp = (struct sockaddr *)&unix_domain_address ;
	addrlen = sizeof (sockaddr_un) ;
    } else {
	fprintf (stderr, "Connect: argument #2 invalid value type\n") ;
	return -1 ;
    }
    st = connect (arg[0].intval, addrp, addrlen) ;
    if (st < 0) {
	return 0 ;
    }
    *retval = 0 ;
    return 0 ;
}

static char acceptClass[] = "\n\
class AcceptInfo { \n\
	public newfd ; \n\
	public sin_port ; \n\
	public client_address ; \n\
} ;\n\
"  ;

//----------------------------------------------------------------
// Listen (socket, backlog) ;
//	description:
//		set a limit for incoming connections.
//	parameters:
//		socket .... a socket descriptor
//		backlog ... number of connections waiting for
//	returns:
//		[success] 0
//		[failure] -1
//----------------------------------------------------------------
static int __Listen (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = -1 ;
    if (args != 2)
	return 0 ;
    if (!arg[0].IsInteger() || !arg[1].IsInteger())
	return 0 ;
    int st ;
    st = listen (arg[0].intval, arg[1].intval) ;
    *retval = st ;
    return 0 ;
}

//----------------------------------------------------------------
//  Accept (socket)
//	description:
//		Accept connection request.
//	parameters:
//		socket ... a socket descriptor.
//	returns:
//		[success] an instance of class AcceptInfo.
//			class AcceptInfo { public newfd,
//					   public sin_port
//					   public client_address ; }
//			member: newfd:
//				new socket descriptor.
//			sin_port:
//				port number  (if INET domain)
//			client_address:
//				ip address of client. (cmm array, 
//				if INET domain)
//		[failure] void
//----------------------------------------------------------------
static int __Accept (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    char addr_buffer[512] ;
    struct sockaddr     *addr ;
    struct sockaddr_in  *in_addr ;
    struct sockaddr_un  *un_addr ;
    int nlen = 512 ;
    int newsock ;
    Instance *ip ;
    Value *retinfo ;

    retval->Init () ;
    if (args != 1)
	return 0 ;
    addr = (struct sockaddr *)addr_buffer ;
    newsock = accept (arg[0].intval, addr, &nlen) ;
    if (newsock < 0) {
	return 0 ;
    }
    ip = cenv->NewInstance ("AcceptInfo") ;
    if (!ip) {
	fprintf (stderr, "Accept : No heap\n") ;
	return -1 ;
    }
    retinfo = cenv->varstack.TmPush (ip) ;
    cenv->SetValue (retinfo, "newfd", newsock) ;
    if (nlen == 0) {
	cenv->SetValue (retinfo, "sin_port", 0) ;
    } else if (addr->sa_family == AF_UNIX) {
	un_addr = (struct sockaddr_un *)addr ;
	char *s ;
	cenv->SetValue (retinfo, "sin_port", 0) ;
	s = cenv->heap_area.AllocString (un_addr->sun_path, &cenv->varstack) ;
	cenv->SetValue (retinfo, "client_address", s) ;
	*retval = s ;
    } else if (addr->sa_family == AF_INET) {
	in_addr = (struct sockaddr_in *)addr ;
	int port = ntohs (in_addr->sin_port) ;
	unsigned long ipaddr = ntohl (in_addr->sin_addr.s_addr) ;

	cenv->SetValue (retinfo, "sin_port", port) ;

	Value inaddr, *inaddrp ;
	inaddr   = cenv->AllocArray (4) ;
	inaddrp = cenv->varstack.TmPush (inaddr) ;
	inaddrp->arr.arrp[0] = (int)((ipaddr >> 24) & 0xff) ;
	inaddrp->arr.arrp[1] = (int)((ipaddr >> 16) & 0xff) ;
	inaddrp->arr.arrp[2] = (int)((ipaddr >>  8) & 0xff) ;
	inaddrp->arr.arrp[3] = (int)(ipaddr & 0xff) ;
	cenv->SetValue (retinfo, "client_address", *inaddrp) ;
	cenv->varstack.TmPop () ;
    } else {
	cenv->SetValue (retinfo, "sin_port", 0) ;
    }
    *retval = cenv->varstack.VPop () ;
    return 0 ;
}

//----------------------------------------------------------------
// GetServerPort (service, protocol) ;
//	description:
//		get a port number of specified service
//	parameters:
//		service .... service name (string).
//		protocol ... protocol name used by specified service.
//				(string)
//	returns:
//		[success] .. a port number.
//		[failure] .. -1
//----------------------------------------------------------------
static int __GetServerPort (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    struct servent *serve ;

    *retval = -1 ;
    if (args != 2)
	return 0 ;
    if (!arg[0].IsString() || !arg[1].IsString())
	return 0 ;
    serve = getservbyname (arg[0].sptr, arg[1].sptr) ;
    if (!serve)
	return 0 ;
    *retval = (int)ntohs(serve->s_port) ;
    return 0 ;
}

void InstallCmmSock (CmmInterp *cmm)
{
    cmm->Execute (acceptClass) ;
    cmm->AddGlobalVar ("AF_UNIX",	 AF_UNIX) ;
    cmm->AddGlobalVar ("AF_INET",	 AF_INET) ;
    cmm->AddGlobalVar ("PF_UNIX",	 PF_UNIX) ;
    cmm->AddGlobalVar ("PF_INET",	 PF_INET) ;
    cmm->AddGlobalVar ("SOCK_STREAM",	 SOCK_STREAM) ;
    cmm->AddGlobalVar ("SOCK_DGRAM",	 SOCK_DGRAM) ;
    cmm->AddGlobalVar ("SOCK_SEQPACKET", SOCK_SEQPACKET) ;
    cmm->AddGlobalVar ("SOCK_RAW",	 SOCK_RAW) ;
    cmm->AddGlobalVar ("SOCK_RDM",	 SOCK_RDM) ;
    cmm->AddBuiltin   ("GetHostName",	 __GetHostName) ;
    cmm->AddBuiltin   ("GetHostAddress", __GetHostAddress) ;
    cmm->AddBuiltin   ("GetServerPort",  __GetServerPort) ;
    cmm->AddBuiltin   ("Socket",	 __Socket) ;
    cmm->AddBuiltin   ("Bind",		 __Bind) ;
    cmm->AddBuiltin   ("Connect",	 __Connect) ;
    cmm->AddBuiltin   ("Accept",	 __Accept) ;
    cmm->AddBuiltin   ("Listen",	 __Listen) ;
}
