//----------------------------------------------------------------
// syscall.cc: UNIX systemcall for Cmm
//----------------------------------------------------------------
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <cmm/cmm.h>
#include <cmm/syscall.h>

#ifdef linux
#define SIGARG int signum
#endif

#ifdef sony_news
#define SIGARG void
#endif

//----------------------------------------------------------------
// getpid ()
//	description:
//		get the process id of the current process
//	parameters:
//		none
//	returns:
//		the process id of the current process.
//----------------------------------------------------------------
static int __getpid (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = (int)getpid () ;
    return 0 ;
}

//----------------------------------------------------------------
// getppid ()
//	description:
//		get process the process ID of the parent of the
//		current process
//	parameters:
//		none
//	returns:
//		the process id of tha parent of the current process.
//----------------------------------------------------------------
static int __getppid (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = (int)getppid () ;
    return 0 ;
}

static int __sleep (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 1)
	return 0 ;
    if (!arg[0].IsInteger())
	return 0 ;
    *retval = (int)sleep ((unsigned int)arg[0].intval) ;
    return 0 ;
}

static CmmInterp *interp[NSIG] = { NULL };
static Value *cmmSigHandlerP   = NULL ; 

static void __cmm_sighandler (int sig)
{
    int sigmask ;
    int lastmask ;

    // block signal to avoid to confuse interpretter
    sigmask = sigmask (sig) ;
    lastmask = sigblock (sigmask) ;
    
    CmmInterp *cmmi = interp[sig] ;
    Value cmmSigHandler = cmmSigHandlerP->arr.arrp[sig] ;
    Value ret, arg ;

//    arg = sig ;
//    printf ("SIGHANDLER === signal = %d\n", sig) ;
//    cmmi->Call (&cmmSigHandler, (Value *)NULL, &arg, 1, &ret) ;
    cmmi->HandleSignal (&cmmSigHandler, sig) ;
    sigsetmask (lastmask) ;
}

static int __alarm (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init ()  ;
    if (args != 1 || !arg[0].IsInteger()) {
	return 0 ;
    }
    *retval = (int)alarm ((long)arg[0].intval) ;
    return 0 ;
}

//----------------------------------------------------------------
// kill (process_id, signumber)
//	descpription:
//		send the signal specified by signumber to the process
//		specieifed by process_id.
//	parameters:
//		process_id  the process id to send the signal to.
//		signumber   the number of signal.
//	returns:
//		[success] 0
//		[falilre] -1
//----------------------------------------------------------------
static int __kill (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 2 || !arg[0].IsInteger() || !arg[1].IsInteger())
	return 0 ;
    *retval = kill ((pid_t)arg[0].intval, arg[1].intval) ;
    return 0 ;
}

static int __pause (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    *retval = pause () ;
    return 0 ;
}

//----------------------------------------------------------------
// signal (signumber, handler) 
//	description:
//		Set a signal handler for signumber.
//	parameter:
//		signumber ... the Signal number (integer)
//		handler ..... the entry of Cmm function.
//			      or, SIG_DFL for the default signal
//			      action or SIG_IGN to ignore signal
//	returns:
//		the last signal hander.
//----------------------------------------------------------------
static int __signal (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 2)
	return 0 ;
    int signo = arg[0].intval ;
    if (signo < 1 || signo >= NSIG)
	return 0 ;
    if (arg[1].IsInteger()) {
	if (arg[1].intval == (int)SIG_DFL) {
	    interp[signo] = NULL ;
	    *retval = cmmSigHandlerP->arr.arrp[signo] ;
	    cmmSigHandlerP->arr.arrp[signo] = (int)SIG_DFL ;
	    signal (signo, SIG_DFL) ;
	}
	if (arg[1].intval == (int)SIG_IGN) {
	    interp[signo] = NULL ;
	    *retval = cmmSigHandlerP->arr.arrp[signo] ;
	    cmmSigHandlerP->arr.arrp[signo] = (int)SIG_IGN ;
	    signal (signo, SIG_IGN) ;
	}
	return 0 ;
    }
    interp[signo] = cenv ;
    *retval = cmmSigHandlerP->arr.arrp[signo] ;
    cmmSigHandlerP->arr.arrp[signo] = arg[1] ;
    signal (signo, __cmm_sighandler) ;
    return 0 ;
}

int __fork (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = fork () ;
    return 0 ;
}

int __vfork (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    *retval = vfork () ;
    return 0 ;
}

static int __Exec (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    int eargs = args ;
    char **exargp ;
    int i ;

    retval->Init () ;
    for (i = 0 ; i < args ; i++)
	if (!arg[i].IsString())
	    return 0 ;
    exargp = (char **)alloca (sizeof(char *) * (args+1)) ;
    for (i = 0 ; i < args ; i++)
	exargp[i] = arg[i].sptr ;
    exargp[i] = NULL ;
    execvp (arg[0].sptr, exargp) ;
    // Not reached if success 
    perror ("Exec") ;
    return 0 ;
}

//----------------------------------------------------------------
// wait ()
//	description:
//		suspends execution of current process while a child is
//		running.
//	parameters:
//		none
//	returns:
//		Cmm wait returns the Cmm array of 2 elements that
//		contains the process id of the exited child, and its
//		status.[pid, status]
//----------------------------------------------------------------
int __wait (CmmInterp *cenv, int args, Variable *arg, Value *retval)
{
    int status ;
    pid_t pid ;

    retval->Init () ;
    pid = wait (&status) ;
    if (pid < 0)
	return 0 ;
    Value r ;
    r = cenv->AllocArray (2) ;
    r.arr.arrp[0] = (int)pid ;
    r.arr.arrp[1] = status ;
    *retval = r ;
    return 0 ;
}

int __waitpid (CmmInterp *cmm, int args, Variable *arg, Value *retval)
{
    retval->Init () ;
    if (args != 2 || !arg[0].IsInteger() || !arg[1].IsInteger())
	return 0 ;
    pid_t pid ;
    int status ;
    pid = waitpid ((pid_t)arg[0].intval, &status, arg[1].intval) ;
    if (pid < 0)
	return 0 ;
    Value r ;
    r = cmm->AllocArray (2) ;
    r.arr.arrp[0] = (int)pid ;
    r.arr.arrp[1] = status ;
    *retval = r ;
    return 0 ;
}

void InstallSysCall (CmmInterp *cmm)
{
    Value sigHandlerArr ;
    
    // Allocate the space for the Cmm Signal Handler Entries.
    // Not to lose the entris when Garbage Collection occured.
    sigHandlerArr = cmm->AllocArray (NSIG) ;
    cmmSigHandlerP = cmm->varstack.Push ("___sighandlers___", sigHandlerArr) ;
    int i ;
    // Initialize all entries to SIG_DFL
    for (i = 0 ; i < NSIG ; i++) {
	cmmSigHandlerP->arr.arrp[i] = (int)SIG_DFL ;
    }
    // Set signal constant
    cmm->AddGlobalVar ("SIG_DFL",	(int)SIG_DFL) ;
    cmm->AddGlobalVar ("SIG_IGN",	(int)SIG_IGN) ;
    cmm->AddGlobalVar ("SIG_ERR",	(int)SIG_ERR) ;
#ifdef SIGHUP
    cmm->AddGlobalVar ("SIGHUP",	SIGHUP) ;
#endif
#ifdef SIGINT
    cmm->AddGlobalVar ("SIGINT",	SIGINT) ;
#endif
#ifdef SIGQUIT
    cmm->AddGlobalVar ("SIGQUIT",	SIGQUIT) ;
#endif
#ifdef SIGILL
    cmm->AddGlobalVar ("SIGILL",	SIGILL) ;
#endif
#ifdef SIGTRAP
    cmm->AddGlobalVar ("SIGTRAP",	SIGTRAP) ;
#endif
#ifdef SIGABRT
    cmm->AddGlobalVar ("SIGABRT",	SIGABRT) ;
#endif
#ifdef SIGIOT
    cmm->AddGlobalVar ("SIGIOT",	SIGIOT) ;
#endif
#ifdef SIGBUS
    cmm->AddGlobalVar ("SIGBUS",	SIGBUS) ;
#endif
#ifdef SIGFPE
    cmm->AddGlobalVar ("SIGFPE",	SIGFPE) ;
#endif
#ifdef SIGKILL
    cmm->AddGlobalVar ("SIGKILL",	SIGKILL) ;
#endif
#ifdef SIGUSR1
    cmm->AddGlobalVar ("SIGUSR1",	SIGUSR1) ;
#endif
#ifdef SIGSEGV
    cmm->AddGlobalVar ("SIGSEGV",	SIGSEGV) ;
#endif
#ifdef SIGUSR2
    cmm->AddGlobalVar ("SIGUSR2",	SIGUSR2) ;
#endif
#ifdef SIGPIPE
    cmm->AddGlobalVar ("SIGPIPE",	SIGPIPE) ;
#endif
#ifdef SIGALRM
    cmm->AddGlobalVar ("SIGALRM",	SIGALRM) ;
#endif
#ifdef SIGTERM
    cmm->AddGlobalVar ("SIGTERM",	SIGTERM) ;
#endif
#ifdef SIGSTKFLT
    cmm->AddGlobalVar ("SIGSTKFLT",	SIGSTKFLT) ;
#endif
#ifdef SIGCHLD
    cmm->AddGlobalVar ("SIGCHLD",	SIGCHLD) ;
#endif
#ifdef SIGCONT
    cmm->AddGlobalVar ("SIGCONT",	SIGCONT) ;
#endif
#ifdef SIGSTOP
    cmm->AddGlobalVar ("SIGSTOP",	SIGSTOP) ;
#endif
#ifdef SIGTSTP
    cmm->AddGlobalVar ("SIGTSTP",	SIGTSTP) ;
#endif
#ifdef SIGTTIN
    cmm->AddGlobalVar ("SIGTTIN",	SIGTTIN) ;
#endif
#ifdef SIGTTOU
    cmm->AddGlobalVar ("SIGTTOU",	SIGTTOU) ;
#endif
#ifdef SIGIO
    cmm->AddGlobalVar ("SIGIO",	SIGIO) ;
#endif
#ifdef SIGPOLL
    cmm->AddGlobalVar ("SIGPOLL",	SIGPOLL) ;
#endif
#ifdef SIGURG
    cmm->AddGlobalVar ("SIGURG",	SIGURG) ;
#endif
#ifdef SIGXCPU
    cmm->AddGlobalVar ("SIGXCPU",	SIGXCPU) ;
#endif
#ifdef SIGXFSZ
    cmm->AddGlobalVar ("SIGXFSZ",	SIGXFSZ) ;
#endif
#ifdef SIGVTALRM
    cmm->AddGlobalVar ("SIGVTALRM",	SIGVTALRM) ;
#endif
#ifdef SIGPROF
    cmm->AddGlobalVar ("SIGPROF",	SIGPROF) ;
#endif
#ifdef SIGWINCH
    cmm->AddGlobalVar ("SIGWINCH",	SIGWINCH) ;
#endif
#ifdef SIGPWR
    cmm->AddGlobalVar ("SIGPWR",	SIGPWR) ;
#endif
#ifdef SIGUNUSED
    cmm->AddGlobalVar ("SIGUNUSED",	SIGUNUSED) ;
#endif

    // flags for wait
    cmm->AddGlobalVar ("WNOHANG",	WNOHANG) ;
    cmm->AddGlobalVar ("WUNTRACED",	WUNTRACED) ;

    // system calls
    cmm->AddBuiltin ("getpid",		__getpid) ;
    cmm->AddBuiltin ("getppid",		__getppid) ;
    cmm->AddBuiltin ("signal",		__signal) ;
    cmm->AddBuiltin ("sleep",		__sleep) ;
    cmm->AddBuiltin ("alarm",		__alarm) ;
    cmm->AddBuiltin ("fork",		__fork) ;
    cmm->AddBuiltin ("vfork",		__vfork) ;
    cmm->AddBuiltin ("kill",		__kill) ;
    cmm->AddBuiltin ("pause",		__pause) ;
    cmm->AddBuiltin ("Exec",		__Exec) ;
    cmm->AddBuiltin ("wait",		__wait) ;
    cmm->AddBuiltin ("waitpid",		__waitpid) ;
}
