;; cmm.el  -- Cmm code editing commands for Emacs
;;
;;  This code is modified from c-mode.el, cplus-md.el
;;  and add some Cmm specific features.
;;

(defvar cmm-mode-map nil
  "Keymap for Cmm Mode")

(defvar cmm-mode-syntax-table nil
  "Syntax table for Cmm Mode")

(defvar cmm-indent-level 4
  "Indentation of Cmm statements with respect to containing block.")

(defvar cmm-tab-always-indent t
  "*Non-nil means TAB in Cmm mode should always reindent the current line,
regardless of where in the line point is when the TAB command is used.")

(defvar cmm-brace-offset -4
  "*Extra indentation for braces, compared with other text in same context.")

(defvar cmm-brace-imaginary-offset 0
  "*Imagined indentation of a Cmm open brace that actually follows a statement.")

(defvar cmm-continued-brace-offset 0
  "*Extra indent for substatements that start with open-braces.
This is in addition to c-continued-statement-offset.")

(defvar cmm-continued-statement-offset 4
  "*Extra indent for lines not starting new statements.")

(defvar cmm-auto-newline nil
  "*Non-nil means automatically newline before and after braces,
and after colons and semicolons, inserted in C code.")

(if cmm-mode-map
    ()
  (setq cmm-mode-map (make-sparse-keymap))
  (define-key cmm-mode-map "\C-j" 'cmm-newline)
  (define-key cmm-mode-map "\C-i" 'cmm-indent-command)
  (define-key cmm-mode-map "}"	  'electric-cmm-brace)
  (define-key cmm-mode-map "{"	  'electric-cmm-brace))

(if cmm-mode-syntax-table
    ()
  (setq cmm-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\\  "\\"    cmm-mode-syntax-table)
  (modify-syntax-entry ?#   "."     cmm-mode-syntax-table)
  (modify-syntax-entry ?/   ". 12"  cmm-mode-syntax-table)
  (modify-syntax-entry ?'   "\""    cmm-mode-syntax-table)
  (modify-syntax-entry ?\n  ">"	    cmm-mode-syntax-table)
  (modify-syntax-entry ?\r  ">"     cmm-mode-syntax-table)
  (modify-syntax-entry ?+ "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?- "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?= "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?% "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?< "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?> "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?& "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?| "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?` "."	    cmm-mode-syntax-table)
  (modify-syntax-entry ?\' "\""	    cmm-mode-syntax-table)
  (modify-syntax-entry ?`  "'"	    cmm-mode-syntax-table)
  (modify-syntax-entry ?@  "'"	    cmm-mode-syntax-table))

(defun cmm-mode ()
  "Major mode for editing Cmm code.
\\{cmm-mode-map}
Variables controlling indentation style:
 cmm-tab-always-indent
    Non-nil means TAB in Cmm mode should always reindent the current line,
    regardless of where in the line point is when the TAB command is used.
 cmm-auto-newline
    Non-nil means automatically newline before and after braces,
    and after colons and semicolons, inserted in C code.
 cmm-indent-level
    Indentation of C statements within surrounding block.
    The surrounding block's indentation is the indentation
    of the line on which the open-brace appears.
 cmm-continued-statement-offset
    Extra indentation given to a substatement, such as the
    then-clause of an if or body of a while.
 cmm-continued-brace-offset
    Extra indentation given to a brace that starts a substatement.
    This is in addition to c-continued-statement-offset.
 cmm-brace-offset
    Extra indentation for line if it starts with an open brace.
 cmm-brace-imaginary-offset
    An open brace following other text is treated as if it were
    this far to the right of the start of its line.

Settings for K&R and BSD indentation styles are
  cmm-indent-level                5    8
  cmm-continued-statement-offset  5    8
  cmm-brace-offset               -5   -8
  cmm-label-offset               -5   -8

Turning on C mode calls the value of the variable c-mode-hook with no args,
if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map cmm-mode-map)
  (setq major-mode 'cmm-mode)
  (setq mode-name "Cmm")
  (set-syntax-table cmm-mode-syntax-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'cmm-indent-line)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'outline-regexp)
  (setq outline-regexp "[^#\n\^M]")
  (make-local-variable 'outline-level)
  (setq outline-level 'cmm-outline-level)
  (make-local-variable 'comment-start)
  (setq comment-start "// ")
  (make-local-variable 'comment-column)
  (setq comment-column 32)
  (make-local-variable 'comment-start-skip)
  (make-local-variable 'comment-indent-function)
  (run-hooks 'cmm-mode-hook))

(defun cmm-newline ()
  (interactive)
  (let (comment-info)
    (if (cmm-within-string)
	(newline)
      (setq comment-info (cmm-comment-line))
      (cmm-indent-command)
      (if comment-info
	  ;; current line is a comment line.
	  ;; newline and indent to the same column of current line
	  (progn
	    (insert "\n")
	    (indent-to (car comment-info))
	    ;;(insert (car (cdr comment-info)))
	    ;;(insert " ")
	    )
	(newline)
	(if cmm-tab-always-indent
	    (save-excursion
	      (beginning-of-line)
	      (cmm-indent-line))
	  (insert "\t"))
	(cmm-indent-command)))))

(defun cmm-indent-line ()
  "Indent current line as Cmm code.
Return the amount the indentation changed by."
  (let ((indent (cmm-calc-indent))
	beg shift-amt
	(case-fold-search nil)
	vardecl
	(pos (- (point-max) (point))))
    (setq vardecl (cmm-in-var-decl))
    (if (not (eq indent 'string))
	(beginning-of-line))
    (setq beg (point))
    (cond ((eq indent nil)
	   (setq indent (current-indentation)))

	  ((eq indent 'string)
	   (insert "\t"))

	  (vardecl
	   (setq indent (+ vardecl 4)))

	  (t
	   (skip-chars-forward " \t")
	   (if (listp indent) (setq indent (car indent)))
	   (cond ((and (looking-at "else\\b")
		       (not (looking-at "else\\s_")))
		  (setq indent (save-excursion
				 (cmm-backward-to-start-of-if)
				 (current-indentation))))
		 
		 ((= (following-char) ?})
		  (setq indent (- indent cmm-indent-level)))
		 
		 ((= (following-char) ?{)
		  (setq indent (+ indent cmm-brace-offset))))))
    (if (eq indent 'string)
	0
      (skip-chars-forward " \t")
      (setq shift-amt (- indent (current-column)))
      (if (zerop shift-amt)
	  (if (> (- (point-max) pos) (point))
	      (goto-char (- (point-max) pos)))
	(delete-region beg (point))
	(indent-to indent)
	;; If initial point was within line's indentation,
	;; position after the indentation.  Else stay at same point in text.
	(if (> (- (point-max) pos) (point))
	    (goto-char (- (point-max) pos))))
      shift-amt)))

(defun cmm-indent-command (&optional whole-exp)
  (interactive "P")
  (if whole-exp
      ;; If arg, always indent this line as C
      ;; and shift remaining lines of expression the same amount.
      (let ((shift-amt (cmm-indent-line))
	    beg end)
	(save-excursion
	  (if cmm-tab-always-indent
	      (beginning-of-line))
	  (setq beg (point))
	  (forward-sexp 1)
	  (setq end (point))
	  (goto-char beg)
	  (forward-line 1)
	  (setq beg (point)))
	(if (> end beg)
	    (indent-code-rigidly beg end shift-amt "#")))
    (if (and (not cmm-tab-always-indent)
	     (save-excursion
	       (skip-chars-backward " \t")
	       (not (bolp))))
	(insert-tab)
      (cmm-indent-line))))

(defun cmm-calc-indent ()
  (interactive)
  (save-excursion
    (beginning-of-line)
    (let ((indent-point (point))
	  (case-fold-search nil)
	  cmmdecl
	  parse-start
	  state
	  containing-sexp)
      (cmm-beginning-of-defun)
      (while (< (point) indent-point)
	(setq parse-start (point))
	(setq state (parse-partial-sexp (point) indent-point))
	(setq containing-sexp (car (cdr state))))

      (cond ((nth 3 state)
	     'string)

	    ((and (nth 4 state) (cmm-comment-line))
	     ;; return nil or t if should not change this line
	     (nth 4 state))

	    ((null containing-sexp)
	     (goto-char indent-point)
	     (skip-chars-forward " \t")
	     (if (= (following-char) ?\{)
		 0
	       (current-indentation)))

	    ((/= (char-after containing-sexp) ?{)
	     (goto-char containing-sexp)
	     (forward-char 1)
	     (skip-chars-forward " \t")
	     (current-column))
	    (t
	     (goto-char indent-point)
	     (cmm-backward-to-noncomment containing-sexp)
	     (if (setq cmmdecl (cmm-decl))
		 (+ cmmdecl
		    (if (memq (preceding-char) '(nil ?\} ?\;))
			0
		      cmm-indent-level))
	       (if (not (memq (preceding-char) '(nil ?\; ?\} ?\{ )))
		   (progn
		     (cmm-backward-to-start-of-continued-exp containing-sexp)
		     (+ cmm-continued-statement-offset (current-column)
			(if (save-excursion (goto-char indent-point)
					    (skip-chars-forward " \t")
					    (eq (following-char) ?{))
			    cmm-continued-brace-offset 0)))
		 (goto-char containing-sexp)
		 (+ (if (and (bolp) (zerop cmm-indent-level))
			(+ cmm-brace-offset cmm-continued-statement-offset)
		      cmm-indent-level)
		    ;; Move back over whitespace before the openbrace.
		    ;; If openbrace is not first nonwhite thing on the line,
		    ;; add the c-brace-imaginary-offset.
		    (progn (skip-chars-backward " \t")
			   (if (bolp) 0 cmm-brace-imaginary-offset))
		    ;; If the openbrace is preceded by a parenthesized exp,
		    ;; move to the beginning of that;
		    ;; possibly a different line
		    (progn
		      (if (eq (preceding-char) ?\))
			  (forward-sexp -1))
		      ;; Get initial indentation of the line we are on.
		      (current-indentation))))))))))

(defun cmm-comment-line ()
  (let (state)
    (setq state (parse-partial-sexp
		 (save-excursion (cmm-beginning-of-defun) (point)) (point)))
    (if (or (null state) (null (nth 4 state)))
	  nil
	(save-excursion
	  (let ((curpos (point))
		ret)
	  (beginning-of-line)
	  (if (null (re-search-forward "\\(#\\|//\\)" curpos 'move))
	      (progn
		(error "not found comment prefix")
		nil)
	    (if (and (= (preceding-char) ?\#)
		     (> (point) 2)
		     (= (char-after (- (point) 2)) ?\$))
		nil
	      (forward-char (if (= (preceding-char) ?/) -2 -1))
	      (list (current-column)
		    (buffer-substring (match-beginning 1) (match-end 1))))))))))

(defun cmm-decl ()
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (if (looking-at "\\(func\\|class\\|method\\|hiden\\)")
	(current-column)
      nil)))
	    
(defun cmm-within-string ()
  (save-excursion
    (nth 3 (parse-partial-sexp (save-excursion (cmm-beginning-of-defun) (point)) (point)))))

(defun cmm-within-string-p (pos)
  (save-excursion
    (goto-char pos)
    (nth 3 (parse-partial-sexp (save-excursion (cmm-beginning-of-defun) (point)) pos))))
  
(defun cmm-indent-level ()
  "Return the indent-level the current statement has.
Do not count labels, case-statements or records."
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (current-column)))

(defun cmm-beginning-of-defun ()
  (if (not (re-search-backward "^\\(var\\|class\\|func\\|method\\|hiden\\)" (point-min) t))
      (goto-char (point-min))))

(defun cmm-backward-to-noncomment (lim)
  (interactive)
  (let (opoint stop)
    (while (not stop)
      (skip-chars-backward " \t\n\r\f" lim)
      (setq opoint (point))
      (cond ((and
	      (re-search-backward "\\(//\\|#\\)"
				  (max (save-excursion (beginning-of-line) (point)) lim)
				  'move)
	      (not (cmm-within-string-p (point)))))
	    (t (beginning-of-line)
	       (setq stop t)
	       (goto-char opoint))))))

(defun electric-cmm-brace (arg)
  "Insert character and correct line's indentation."
  (interactive "P")
  (let (insertpos)
    (if (and (not arg)
	     (eolp)
	     (or (save-excursion
		   (skip-chars-backward " \t")
		   (bolp))
		 (if cmm-auto-newline (progn (cmm-indent-line) (newline) t))))
	(progn
	  (insert last-command-char)
	  (cmm-indent-line)
	  (if cmm-auto-newline
	      (progn
		(newline)
		;; (newline) may have done auto-fill
		(setq insertpos (- (point) 2))
		(cmm-indent-line)))
	  (save-excursion
	    (if insertpos (goto-char (1+ insertpos)))
	    (delete-char -1))))
    (if insertpos
	(save-excursion
	  (goto-char insertpos)
	  (self-insert-command (prefix-numeric-value arg)))
      (self-insert-command (prefix-numeric-value arg)))))

(defun cmm-backward-to-start-of-continued-exp (lim)
  (if (memq (preceding-char) '(?\) ?\"))
      (forward-sexp -1))
  (beginning-of-line)
  (if (<= (point) lim)
      (goto-char (1+ lim)))
  (skip-chars-forward " \t"))

(defun cmm-backward-to-start-of-if (&optional limit)
  "Move to the start of the last \"unbalanced\" `if'."
  (or limit (setq limit (save-excursion (beginning-of-defun) (point))))
  (let ((if-level 1)
	(case-fold-search nil))
    (while (and (not (bobp)) (not (zerop if-level)))
      (backward-sexp 1)
      (cond ((looking-at "else\\b")
	     (setq if-level (1+ if-level)))
	    ((looking-at "if\\b")
	     (setq if-level (1- if-level)))
	    ((< (point) limit)
	     (setq if-level 0)
	     (goto-char limit))))))

(defun cmm-outline-level ()
  (save-excursion
    (skip-chars-forward "\t ")
    (current-column)))


(defun test-cmm-invar-decl ()
  (interactive)
  (print (cmm-in-var-decl)))

(defun cmm-in-var-decl ()
  (save-excursion
    (let ((pos (point))
	  (firstlinep nil)
	  varpos)
      (save-excursion
	(beginning-of-line)
	(skip-chars-forward " \t")
	(if (looking-at "var[^a-zA-Z_]")
	    (setq firstlinep t)))
      (cond (firstlinep
	     nil)
	    ((and (re-search-backward "var[^a-zA-Z_]"
				      (save-excursion (cmm-beginning-of-defun) (point)) t)
		  (not (cmm-within-string))
		  (not (cmm-comment-line)))
	     (setq varpos (point))
	     (if (and (re-search-forward ";" (1+ pos) t)
		      (< (point) pos)
		      (not (cmm-within-string))
		      (not (cmm-comment-line)))
		 nil
	       (goto-char varpos)
	       (current-column)))
	    
	     (t nil)))))

		

		   

	
    